/*-
 * Copyright (c) 2006, 2011 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <libgen.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>
#include <time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>

#include "threadpool.h"
#include "preforkserver.h"
#include "scgi.h"
#include "wsgi.h"
#include "version.h"

/* Some arbitrary constants */
#define MAX_SOCKETS 10
#define DEFAULT_MAX_THREADS INT_MAX
#define DEFAULT_MIN_THREADS 1
#define DEFAULT_MAX_PROCESSES 16
#define DEFAULT_MIN_PROCESSES 1
#define DEFAULT_MAX_IDLE 300
#define REAP_INTERVAL 2

/* Tweakables from glue */
extern int unquoteURL;
extern int multiprocess;
extern int selectEnvHandler(const char *name);

/* Misc UI stuff */
static char *progName;
static int verbosity = 0;
static int forking = 0;

/* Socket handling */
static int socks[MAX_SOCKETS];
static int sockCount, nfds;

/* Threading support */
static ThreadPoolContext *tpCtxt;
static pthread_t _masterThread;

/* Forking support */
static PreforkContext *pfCtxt;
static pid_t _masterProcess;
static int maxProcesses = DEFAULT_MAX_PROCESSES,
  minProcesses = DEFAULT_MIN_PROCESSES;

/* Max idle for worker threads/child processes */
static time_t maxIdle = DEFAULT_MAX_IDLE;

/* Used to hold data about each connection */
struct _ClientData {
  struct _ClientData *next, *prev;
  int sock;
  union {
    struct sockaddr_in in;
    struct sockaddr_in6 in6;
  } addr;
};

/* Overflow queue when forking */
static struct _ClientData overflowHead, overflowTail;

/* Signal handling */
#define MAIN_LOOP_OK 0
#define MAIN_LOOP_DIE 1
#define MAIN_LOOP_RELOAD 2

static volatile sig_atomic_t mainLoopState = MAIN_LOOP_OK;

/* Simple wrapper around strerror */
static void
myperror(const char *subModule)
{
  if (subModule == NULL)
    fprintf(stderr, "%s: %s\n", progName, strerror(errno));
  else
    fprintf(stderr, "%s: %s: %s\n", progName, subModule, strerror(errno));
}

static const char *
sockaddr_to_string(struct sockaddr *sa, char *dst, size_t size)
{
  char addr[INET6_ADDRSTRLEN];
  in_port_t port;

  switch (sa->sa_family) {
  case AF_INET:
    inet_ntop(AF_INET, &((struct sockaddr_in *)sa)->sin_addr, addr,
	      sizeof(addr));
    port = ntohs(((struct sockaddr_in *)sa)->sin_port);
    break;
  case AF_INET6:
    inet_ntop(AF_INET6, &((struct sockaddr_in6 *)sa)->sin6_addr, addr,
	      sizeof(addr));
    port = ntohs(((struct sockaddr_in6 *)sa)->sin6_port);
    break;
  default:
    strncpy(dst, "[unknown address family]", size - 1);
    dst[size - 1] = '\0';
    return dst;
  }

  snprintf(dst, size, "%s:%d", addr, port);
  return dst;
}

/* Connection handler for threads */
static void
threadClientHandler(void *clientData)
{
  struct _ClientData *cd = clientData;
  SCGIContext *ctxt;
  char addr[100];

  if (verbosity)
    fprintf(stderr, "%s: Connection opened (%s)\n",
	    progName,
	    sockaddr_to_string((struct sockaddr *)&cd->addr, addr,
			       sizeof(addr)));
  if ((ctxt = scgiCreateContext(cd->sock,
			       (scgi_handler_t)wsgiHandler,
			       (scgi_body_handler_t)wsgiBodyHandlerNC))) {
    int nfds = cd->sock + 1;

    /* Main loop */
    while (mainLoopState == MAIN_LOOP_OK) {
      fd_set readfds;
      int ready;

      /* Wait on client socket */
      FD_ZERO(&readfds);
      FD_SET(cd->sock, &readfds);

      if ((ready = select(nfds, &readfds, NULL, NULL, NULL)) == -1 &&
	  errno != EINTR) {
	myperror("select");
	exit(2);
      }

      if (ready > 0) {
	if (FD_ISSET(cd->sock, &readfds)) {
	  /* Handle request from SCGI client */
	  if (scgiProcessInput(ctxt) != SCGI_OK)
	    break;
	}
      }
    }

    scgiDestroyContext(ctxt);
  }
  if (verbosity)
    fprintf(stderr, "%s: Connection closed (%s)\n",
	    progName,
	    addr);

  free(cd);
}

/* Connection handler for processes */
static void
preforkHandler(int parent, int sock)
{
  SCGIContext *ctxt;
  char addr[100];

  if (verbosity) {
    union {
      struct sockaddr_in in;
      struct sockaddr_in6 in6;
    } sa;
    socklen_t salen = sizeof(sa);

    /* Get client's address (again) */
    if (!getpeername(sock, (struct sockaddr *)&sa, &salen)) {
      sockaddr_to_string((struct sockaddr *)&sa, addr, sizeof(addr));
    }
    else
      strcpy(addr, "[getpeername failed]");

    fprintf(stderr, "%s: Connection opened (%s)\n",
	    progName,
	    addr);
  }
  if ((ctxt = scgiCreateContext(sock,
			       (scgi_handler_t)wsgiHandler,
			       (scgi_body_handler_t)wsgiBodyHandlerNC))) {
    int nfds = sock;

    /* Figure out max number of FDs we have to check */
    if (parent > nfds)
      nfds = parent;
    nfds++;

    /* Main loop */
    while (mainLoopState == MAIN_LOOP_OK) {
      fd_set readfds;
      int ready;

      /* Wait on both client and parent sockets */
      FD_ZERO(&readfds);
      FD_SET(sock, &readfds);
      FD_SET(parent, &readfds);

      if ((ready = select(nfds, &readfds, NULL, NULL, NULL)) == -1 &&
	  errno != EINTR) {
	myperror("select");
	exit(2);
      }

      if (ready > 0) {
	if (FD_ISSET(sock, &readfds)) {
	  /* Handle request from SCGI client */
	  if (scgiProcessInput(ctxt) != SCGI_OK)
	    break;
	}
	if (FD_ISSET(parent, &readfds)) {
	  /* Only thing this could mean is that the parent died */
	  break;
	}
      }
    }

    scgiDestroyContext(ctxt);
  }
  if (verbosity)
    fprintf(stderr, "%s: Connection closed (%s)\n",
	    progName,
	    addr);
}

/* Connection handoff for threads */
static void
threadHandoff(struct _ClientData *cd)
{
  if (threadpool_run(tpCtxt, threadClientHandler, cd)) {
    myperror("threadpool_run");
    exit(2);
  }
}

/* Remove queued jobs, if any, and try to hand them off again */
static void
process_overflow(int freed)
{
  int i;

  for (i = 0; i < freed; i++) {
    struct _ClientData *cd = overflowHead.next;
    if (cd->next) {
      /* Got one, try to run it */
      if (prefork_run(pfCtxt, cd->sock)) {
	if (errno == EAGAIN) break; /* Unexpected, but not an error */
	myperror("prefork_run");
	exit(2);
      }

      /* Successfully sent on its way, dequeue it */
      cd->prev->next = cd->next;
      cd->next->prev = cd->prev;

      /* And clean it up */
      close(cd->sock);
      free(cd);
    }
    else
      break; /* Queue empty */
  }
}

/* Connection handoff for processes */
static void
procHandoff(struct _ClientData *cd)
{
  while (prefork_run(pfCtxt, cd->sock)) {
    if (errno == EAGAIN) {
      /* No available children. Try to spawn another. */
      if (prefork_child_count(pfCtxt) < maxProcesses) {
	if (prefork_new_child(pfCtxt, 0)) { /* Only a temporary child */
	  myperror("prefork_new_child");
	  exit(2);
	}
	/* Attempt prefork_run() again... */
      }
      else {
	/* Reached the limit, queue it up */
	cd->next = &overflowTail;
	cd->prev = overflowTail.prev;

	overflowTail.prev->next = cd;
	overflowTail.prev = cd;
	return; /* NB don't close socket yet */
      }
    }
    else {
      myperror("prefork_run");
      exit(2);
    }
  }
  close(cd->sock);
  free(cd);
}

/* Creates the listening socket(s) */
static int
createSocket(const char *ifname, int port)
{
  struct addrinfo hints, *res, *res0;
  int error;
  int s;
  int on = 1;
  char portBuf[10];

  snprintf(portBuf, sizeof(portBuf), "%d", port);

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;
  error = getaddrinfo(ifname, portBuf, &hints, &res0);
  if (error) {
    fprintf(stderr, "%s: getaddrinfo: %s\n", progName,
	    gai_strerror(error));
    return -1;
  }

  sockCount = 0;
  nfds = -1;
  for (res = res0; res; res = res->ai_next) {
    if ((s = socket(res->ai_family, res->ai_socktype,
		    res->ai_protocol)) == -1) {
      myperror("socket");
      continue;
    }

    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) {
      myperror("setsockopt");
      close(s);
      continue;
    }

    if (bind(s, res->ai_addr, res->ai_addrlen)) {
      myperror("bind");
      close(s);
      continue;
    }

    if (listen(s, SOMAXCONN)) {
      myperror("listen");
      close(s);
      continue;
    }

    socks[sockCount++] = s;

    if (s > nfds)
      nfds = s;

    if (sockCount == MAX_SOCKETS) {
      fprintf (stderr, "%s: Maximum number of sockets bound\n",
	       progName);
      break;
    }
  }
  freeaddrinfo(res0);

  nfds++;

  return sockCount ? 0 : -1;
}

/* Signal handler for SIGHUP/SIGINT/SIGTERM */
static void
die(int sig)
{
  mainLoopState = sig == SIGHUP ? MAIN_LOOP_RELOAD : MAIN_LOOP_DIE;

  /* Relay signal to parent process, in case we're backgrounded */
  if (forking) {
    if (getpid() != _masterProcess)
      kill(_masterProcess, sig);
  }
}

/* Displays command-line usage */
static void
usage(void)
{
  int len, i;
  char *filler, *c;

  len = strlen(progName);
  if ((filler = malloc(len + 1)) == NULL) {
    myperror("malloc");
    exit(2);
  }
  for (i = 0, c = filler; i < len; i++)
    *(c++) = ' ';
  *c = '\0';

  fprintf(stderr,
	  "Usage: %s [-BFQVv] [-l <logFile>]\n"
	  "       %s [-h <ifname>] [-p <port>]\n"
	  "       %s [-n <maxConnections>] [-m <minConnections>]"
	  " [-i <maxIdle>]\n"
	  "       %s [-E pass-through|request-uri|script-name]\n"
	  "       %s <moduleName> <appName> [<scriptName>]\n",
	  progName, filler, filler, filler, filler);
  exit(1);
}

int
main(int argc, char *argv[])
{
  int ch;
  char *end;
  char *scriptName = "", *ifname = "localhost";
  int port = 4000;
  int daemonize = 0;
  char *logName = NULL;
  int minThreads = DEFAULT_MIN_THREADS, maxThreads = DEFAULT_MAX_THREADS;
  void (*handoff)(struct _ClientData *);

  if ((progName = strdup(basename(argv[0]))) == NULL) {
    perror(NULL);
    exit(2);
  }

  /* Command-line parsing */
  while ((ch = getopt(argc, argv, "BE:FQVl:h:i:m:n:p:t:v")) != -1) {
    switch (ch) {
    case 'B':
      daemonize++;
      break;
    case 'E':
      if (selectEnvHandler(optarg)) {
	fprintf (stderr, "%s: bad environment profile name\n", progName);
	exit(1);
      }
      break;
    case 'F':
      forking++;
      break;
    case 'Q':
      unquoteURL = 0;
      break;
    case 'V':
      fprintf (stderr, "%s %s (Python %s)\n", progName, VERSION_STRING,
	       wsgiPyVersion);
      exit(1);
      break;
    case 'l':
      logName = optarg;
      break;
    case 'h':
      ifname = optarg;
      break;
    case 'i':
      maxIdle = strtol(optarg, &end, 10);
      if (*end || maxIdle < 0) {
	fprintf(stderr, "%s: Bad maxIdle argument\n", progName);
	exit(1);
      }
      break;
    case 't':
      fprintf(stderr, "%s: -t option is deprecated, use -n instead\n", progName);
      /* FALLTHRU */
    case 'n':
      maxThreads = strtol(optarg, &end, 10);
      if (*end || maxThreads < 1) {
	fprintf(stderr, "%s: Bad maxConnections argument\n", progName);
	exit(1);
      }
      maxProcesses = maxThreads;
      break;
    case 'm':
      minThreads = strtol(optarg, &end, 10);
      if (*end || minProcesses < 0) {
	fprintf(stderr, "%s: Bad minConnections argument\n", progName);
	exit(1);
      }
      minProcesses = minThreads;
      break;
    case 'p':
      port = strtol(optarg, &end, 10);
      if (*end || port < 1 || port > 65535) {
	fprintf (stderr, "%s: Bad port argument\n", progName);
	exit(1);
      }
      break;
    case 'v':
      verbosity++;
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (argc < 2 || argc > 3)
    usage();

  if (argc > 2) {
    /* Validate scriptName */
    scriptName = argv[2];
    if (scriptName[0]) {
      if (scriptName[0] != '/') {
	fprintf(stderr, "%s: scriptName must start with '/'\n", progName);
	exit(1);
      }
      if (scriptName[strlen(scriptName) - 1] == '/') {
	fprintf(stderr, "%s: scriptName must not end with '/'\n", progName);
	exit(1);
      }
    }
  }

  if (logName != NULL) {
    /* Set up log file */
    int log;

    if ((log = open(logName, O_CREAT|O_APPEND|O_WRONLY, 0666)) == -1) {
      myperror(logName);
      exit(2);
    }

    dup2(log, STDERR_FILENO);

    if (log != STDERR_FILENO)
      close(log);
  }

  if (daemonize) {
    /* Do this ourselves so we have more control */
    int fd = open("/dev/null", O_RDWR);
    dup2(fd, STDIN_FILENO);
    dup2(fd, STDOUT_FILENO);
    if (logName == NULL)
      dup2(fd, STDERR_FILENO);
    if (fd > STDERR_FILENO)
      close(fd);

    if (daemon (1, 1)) {
      myperror("daemon");
      exit(2);
    }
  }

  /* Create listening sockets */
  if (createSocket(ifname, port))
    exit(2);

  /* Initialize Python/WSGI */
  if (wsgiInit(argv[0], argv[1], scriptName, progName))
    exit(2);

  /* Ensure Python interpreter is shut down on exit */
  atexit(wsgiCleanup);

  /* Do things a little differently depending on whether we use threads or
     processes */
  if (!forking) {
    if ((tpCtxt = threadpool_init(minThreads, maxThreads, maxIdle)) == NULL) {
      myperror("threadpool_init");
      exit(2);
    }

    _masterThread = pthread_self();
    handoff = threadHandoff;
  }
  else {
    if ((pfCtxt = prefork_init(preforkHandler, maxIdle)) == NULL) {
      myperror("prefork_init");
      exit(2);
    }

    /* Initialize overflow queue */
    overflowHead.next = &overflowTail;
    overflowHead.prev = NULL;
    overflowTail.next = NULL;
    overflowTail.prev = &overflowHead;

    _masterProcess = getpid();
    handoff = procHandoff;
  }

  /* for wsgi.multithread & wsgi.multiprocess */
  multiprocess = forking;

  /* Install signal handlers */
  signal(SIGHUP, die);
  signal(SIGINT, die);
  signal(SIGTERM, die);

  /* Main loop */
  while (mainLoopState == MAIN_LOOP_OK) {
    fd_set readfds;
    int i;
    int pf_nfds = 0;
    struct timeval timeout;
    int ready;

    /* Wait on all listening sockets */
    FD_ZERO(&readfds);
    for (i = 0; i < sockCount; i++)
      FD_SET(socks[i], &readfds);

    /* Special handling for prefork server */
    if (pfCtxt != NULL) {
      /* Ensure minimum number of child processes */
      while (prefork_child_count(pfCtxt) < minProcesses) {
	if (prefork_new_child(pfCtxt, 1)) { /* These should be eternal */
	  myperror("prefork_new_child");
	  exit(2);
	}
      }

      /* Fill fd_set with all child descriptors */
      pf_nfds = prefork_fill_fd_set(pfCtxt, &readfds);
    }

    /* Wait for activity */
    timeout.tv_sec = REAP_INTERVAL;
    timeout.tv_usec = 0;
    if ((ready = select(pf_nfds > nfds ? pf_nfds : nfds, &readfds, NULL, NULL,
			&timeout)) == -1 && errno != EINTR) {
      myperror("select");
      exit(2);
    }

    /* Scan for active sockets */
    if (ready >= 0) {
      for (i = 0; i < sockCount; i++) {
	if (FD_ISSET(socks[i], &readfds)) {
	  struct _ClientData *cd;
	  socklen_t addrLen = sizeof(cd->addr);

	  if ((cd = calloc(1, sizeof(*cd))) == NULL) {
	    myperror("malloc");
	    exit(2);
	  }

	  /* Accept client connection... */
	  if ((cd->sock = accept(socks[i], (struct sockaddr *)&cd->addr,
				 &addrLen)) != -1) {
	    /* ...and hand it off */
	    handoff(cd);
	  }
	  else if (errno != EINTR) {
	    myperror("accept");
	    free(cd);
	  }
	}
      }

      if (pfCtxt != NULL) {
	int freed;

	if ((freed = prefork_process_children(pfCtxt, &readfds)))
	  process_overflow(freed);
      }
    }
  }

  if (pfCtxt != NULL)
    prefork_shutdown(pfCtxt);
  if (tpCtxt != NULL)
    threadpool_shutdown(tpCtxt);

  if (mainLoopState == MAIN_LOOP_RELOAD) {
    if (verbosity)
      fprintf(stderr, "%s: Exiting for reload...\n", progName);
    exit(3);
  }
  else {
    if (verbosity)
      fprintf(stderr, "%s: Exiting...\n", progName);
    exit(0);
  }
}
