/*-
 * Copyright (c) 2011 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "preforkserver.h"

/* Size of hash table used to keep track of children */
#define CHILDREN_HASH_TABLE_SIZE 337

/* Single octet messages children send to the parent */
#define CHILD_MSG_BUSY 0xff
#define CHILD_MSG_FREE 0x00

/* The state of a child, according to the parent */
enum _ChildState {
  CHILD_AVAILABLE,
  CHILD_FD_SENT,
  CHILD_BUSY
};

/* Data about a child process */
struct _ChildData {
  struct _ChildData *next;
  pid_t pid;
  int fd; /* parent's end, used for sending client fd */
  enum _ChildState state;
};

/* Keep track of child within our table */
static int
track_child(PreforkContext *ctxt, pid_t pid, int fd)
{
  struct _ChildData *cd;
  int key;

  /* Allocate _ChildData first */
  if ((cd = malloc(sizeof(*cd))) == NULL)
    return -1;

  cd->pid = pid;
  cd->fd = fd;
  cd->state = CHILD_AVAILABLE;

  /* Insert into hash table */
  key = pid % CHILDREN_HASH_TABLE_SIZE;
  cd->next = ctxt->children[key];
  ctxt->children[key] = cd;

  return 0;
}

/* Clean up and free a _ChildData structure */
static void
_forget_child(struct _ChildData *cd)
{
    close(cd->fd);
    free(cd);
}

/* Stop tracking child */
static int
forget_child(PreforkContext *ctxt, pid_t pid)
{
  int key = pid % CHILDREN_HASH_TABLE_SIZE;
  struct _ChildData *cd = ctxt->children[key], *prev = NULL;

  while (cd != NULL) {
    if (cd->pid == pid) break;

    prev = cd; /* Track previous child since we're deleting */
    cd = cd->next;
  }

  if (cd != NULL) {
    if (prev != NULL)
      prev->next = cd->next;
    else
      ctxt->children[key] = NULL;

    _forget_child(cd);

    return 0;
  }
  else {
    errno = ESRCH;
    return -1;
  }
}

/* Send a descriptor over a UNIX domain socket */
static int
write_fd(int fd, void *ptr, size_t bytes, int sendfd)
{
  struct msghdr msg;
  struct iovec iov[1];

  union {
    struct cmsghdr cm;
    char control[CMSG_SPACE(sizeof(int))];
  } control_un;
  struct cmsghdr *cmptr;

  memset(&msg, 0, sizeof(msg));
  msg.msg_control = control_un.control;
  msg.msg_controllen = sizeof(control_un.control);

  cmptr = CMSG_FIRSTHDR(&msg);
  cmptr->cmsg_len = CMSG_LEN(sizeof(int));
  cmptr->cmsg_level = SOL_SOCKET;
  cmptr->cmsg_type = SCM_RIGHTS;
  *((int *)CMSG_DATA(cmptr)) = sendfd;

  msg.msg_name = NULL;
  msg.msg_namelen = 0;

  iov[0].iov_base = ptr;
  iov[0].iov_len = bytes;
  msg.msg_iov = iov;
  msg.msg_iovlen = 1;

  return sendmsg(fd, &msg, 0);
}

int
prefork_run(PreforkContext *ctxt, int fd)
{
  int i;
  static char msg[6] = { 'p', 'l', 'e', 'a', 's', 'e' };

  /* Scan for first available child */
  for (i = 0; i < CHILDREN_HASH_TABLE_SIZE; i++) {
    struct _ChildData *cd;
    for (cd = ctxt->children[i]; cd != NULL; cd = cd->next) {
      if (cd->state == CHILD_AVAILABLE) {
	/* Found one */
	ssize_t sent;
	while ((sent = write_fd(cd->fd, msg, sizeof(msg), fd)) == -1 &&
	       errno == EAGAIN) {
	  /* For the rare memoments when the socket would block */
	  fd_set writefds;
	  FD_ZERO(&writefds);
	  FD_SET(cd->fd, &writefds);
	  if (select(cd->fd + 1, NULL, &writefds, NULL, NULL) == -1) break;
	}
	if (sent != sizeof(msg))
	  return -1;
	cd->state = CHILD_FD_SENT;
	close(fd);
	return 0;
      }
    }
  }

  errno = EAGAIN;
  return -1;
}

void
prefork_shutdown(PreforkContext *ctxt)
{
  int i;

  /* Close all sockets to children, signalling death */
  for (i = 0; i < CHILDREN_HASH_TABLE_SIZE; i++) {
    struct _ChildData *cd = ctxt->children[i];
    while (cd != NULL) {
      struct _ChildData *next = cd->next;
      _forget_child(cd);
      cd = next;
    }
  }

  free (ctxt->children);
  free (ctxt);
}

int
prefork_child_count(PreforkContext *ctxt)
{
  return ctxt->childCount;
}

int
prefork_fill_fd_set(PreforkContext *ctxt, fd_set *fds)
{
  int nfds = -1;
  int i;

  for (i = 0; i < CHILDREN_HASH_TABLE_SIZE; i++) {
    struct _ChildData *cd;
    for (cd = ctxt->children[i]; cd != NULL; cd = cd->next) {
      /* Remember largest descriptor */
      if (cd->fd > nfds)
	nfds = cd->fd;
      FD_SET(cd->fd, fds);
    }
  }

  return nfds + 1;
}

int
prefork_process_children(PreforkContext *ctxt, fd_set *fds)
{
  int result = 0;
  int i;
  pid_t pid;

  if (fds != NULL) {
    for (i = 0; i < CHILDREN_HASH_TABLE_SIZE; i++) {
      struct _ChildData *cd;
      for (cd = ctxt->children[i]; cd != NULL; cd = cd->next) {
	if (FD_ISSET(cd->fd, fds)) {
	  ssize_t len;
	  uint8_t buf;

	  if ((len = recv(cd->fd, &buf, sizeof(buf), 0)) == 1) {
	    /* Set child state accordingly */
	    if (buf == CHILD_MSG_BUSY)
	      cd->state = CHILD_BUSY;
	    else if (buf == CHILD_MSG_FREE) {
	      cd->state = CHILD_AVAILABLE;
	      result++;
	    }
	  }
	  else {
	    /* Do nothing for now. Might be EAGAIN, might be something else.
	       Could be EOF too. Regardless, let reaper take care of it. */
	  }
	}
      }
    }
  }

  /* Reap children */
  while ((pid = waitpid(-1, NULL, WNOHANG)) > 0) {
    /* Child died */
    if (!forget_child(ctxt, pid))
      ctxt->childCount--;
  }

  return result;
}

/* Receive a descriptor through a UNIX domain socket */
static ssize_t
read_fd(int fd, void *ptr, size_t bytes, int *recvfd)
{
  struct msghdr msg;
  struct iovec iov[1];
  ssize_t n;

  union {
    struct cmsghdr cm;
    char control[CMSG_SPACE(sizeof(int))];
  } control_un;
  struct cmsghdr *cmptr;

  memset(&msg, 0, sizeof(msg));
  msg.msg_control = control_un.control;
  msg.msg_controllen = sizeof(control_un.control);
  msg.msg_name = NULL;
  msg.msg_namelen = 0;

  iov[0].iov_base = ptr;
  iov[0].iov_len = bytes;
  msg.msg_iov = iov;
  msg.msg_iovlen = 1;

  *recvfd = -1;

  if ((n = recvmsg(fd, &msg, 0)) <= 0)
    return n;

  if ((cmptr = CMSG_FIRSTHDR(&msg)) != NULL &&
      cmptr->cmsg_len == CMSG_LEN(sizeof(int))) {
    if (cmptr->cmsg_level == SOL_SOCKET &&
	cmptr->cmsg_type == SCM_RIGHTS) {
      *recvfd = *((int *)CMSG_DATA(cmptr));
    }
  }

  return n;
}

/* Send a status update to the parent */
static int
child_send_update(int parent, int available)
{
  ssize_t sent;
  uint8_t msg = available ? CHILD_MSG_FREE : CHILD_MSG_BUSY;

  while ((sent = send(parent, &msg, 1, 0)) == -1 && errno == EAGAIN) {
    /* Socket is non-blocking, so have to consider EAGAIN */
    fd_set writefds;
    FD_ZERO(&writefds);
    FD_SET(parent, &writefds);
    if (select(parent + 1, NULL, &writefds, NULL, NULL) == -1) break;
  }

  return sent == 1 ? 0 : -1;
}

/* Main loop for a child process */
static void
child_main(PreforkContext *ctxt, int parent, int eternal)
{
  for (;;) {
    fd_set readfds;
    struct timeval timeout;

    /* Wait on parent socket */
    FD_ZERO(&readfds);
    FD_SET(parent, &readfds);

    timeout.tv_sec = ctxt->maxIdle;
    timeout.tv_usec = 0;
    if (select(parent + 1, &readfds, NULL, NULL,
	       eternal ? NULL : &timeout) != 1) {
      /* If it's -1, it's most likely EINTR */
      /* If it's 0, then this child has expired */
      break;
    }

    if (FD_ISSET(parent, &readfds)) {
      int client;
      char buf[6];

      /* Receive client descriptor from parent */
      if (read_fd(parent, buf, sizeof(buf), &client) == sizeof(buf) &&
	  client >= 0) {
	/* Tell parent we're busy */
	if (child_send_update(parent, 0)) break;

	/* Run handler */
	ctxt->handler(parent, client);

	/* Tell parent we're free again */
	if (child_send_update(parent, 1)) break;
      }
      else
	break; /* Parent most likely dead */
    }
  }

  exit(0);
}

/* Set non-blocking flag on descriptor */
static int
setNonBlocking(int fd)
{
  int flags;

  if ((flags = fcntl(fd, F_GETFL)) == -1)
    return -1;
  flags |= O_NONBLOCK;
  if (fcntl(fd, F_SETFL, flags) == -1)
    return -1;
  return 0;
}

/* Set close-on-exec flag on descriptor */
static int
setCloseOnExec(int fd)
{
  int flags;

  if ((flags = fcntl(fd, F_GETFD)) == -1)
    return -1;
  flags |= FD_CLOEXEC;
  if (fcntl(fd, F_SETFD, flags) == -1)
    return -1;
  return 0;
}

int
prefork_new_child(PreforkContext *ctxt, int eternal)
{
  int socks[2];
  pid_t pid;
  int fd, fd_max;

  /* Create socket pair for parent/child communication */
  if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, socks))
    return -1;

  /* Set close-on-exec on child's socket and make both sockets non-blocking */
  if (setCloseOnExec(socks[1]) || setNonBlocking(socks[0]) ||
      setNonBlocking(socks[1])) {
    close(socks[1]);
    close(socks[0]);
    return -1;
  }

  /* Create child process */
  switch ((pid = fork())) {
  case -1: /* Error */
    close(socks[1]);
    close(socks[0]);
    return -1;

  default: /* Parent */
    close(socks[1]);
    if (track_child(ctxt, pid, socks[0])) {
      close(socks[0]); /* Spawned child will die off */
      return -1;
    }      
    ctxt->childCount++;
    return 0;

  case 0: /* Child */
    /* Close extraneous files */
    for (fd = STDERR_FILENO + 1, fd_max = getdtablesize(); fd < fd_max; fd++) {
      if (fd != socks[1])
	close(fd);
    }

    child_main(ctxt, socks[1], eternal);
    _exit(0);
    /* NOTREACHED */
  }
}

PreforkContext *
prefork_init(prefork_handler_t handler, time_t maxIdle)
{
  PreforkContext *ctxt;

  if ((ctxt = calloc(1, sizeof(*ctxt))) == NULL)
    goto bad;

  ctxt->handler = handler;
  ctxt->maxIdle = maxIdle;;

  /* Allocate hash table */
  if ((ctxt->children = calloc(CHILDREN_HASH_TABLE_SIZE,
			       sizeof(*ctxt->children))) == NULL)
    goto bad;

  /* Can be considered fully initialized at this point */
  return ctxt;

 bad:
  if (ctxt != NULL) {
    if (ctxt->children != NULL) free(ctxt);
    free(ctxt);
  }

  return NULL;
}
