/*-
 * Copyright (c) 2006 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <Python.h>

#include <assert.h>

#include "scgi.h"
#include "wsgi-int.h"

int unquoteURL = 1;
int multiprocess = 0;

typedef int (*envHandler)(Request *self, const char *requestUri,
			  const char *scriptName, const char *pathInfo,
			  const char *givenScriptName, int givenScriptNameLen);

/* The various SCRIPT_NAME/PATH_INFO/REQUEST_URI handlers */

/* Use original values */
static int
passThroughHandler(Request *self, const char *requestUri,
		   const char *scriptName, const char *pathInfo,
		   const char *givenScriptName, int givenScriptNameLen)
{
  int result = -1;
  const char *unqReqUri = NULL;

  if (unquoteURL) {
    if ((unqReqUri = wsgiUnquote(requestUri)) == NULL) {
      PyErr_NoMemory();
      goto out;
    }

    if (wsgiPutEnv(self, "REQUEST_URI", unqReqUri))
      goto out;
  }
  else {
    /* Do nothing */
  }

  result = 0;

 out:
  if (unqReqUri) PyMem_Free((void *)unqReqUri);

  return result;
}

/* Set SCRIPT_NAME/PATH_INFO from REQUEST_URI */
static int
setScriptNamePathInfo(Request *self, const char *requestUri,
		      const char *scriptName, int scriptNameLen)
{
  const char *pathInfo;

  /* SCRIPT_NAME/PATH_INFO */
  pathInfo = requestUri;
  if (scriptNameLen) {
    if (!strncmp(pathInfo, scriptName, scriptNameLen))
      pathInfo = &pathInfo[scriptNameLen];
    else
      fprintf (stderr, "WARNING: SCRIPT_NAME does not match REQUEST_URI\n");
  }

  if (wsgiPutEnv(self, "PATH_INFO", pathInfo))
    return -1;

  if (wsgiPutEnv(self, "SCRIPT_NAME", scriptName))
    return -1;

  return 0;
}

/* Only trust REQUEST_URI */
static int
requestUriHandler(Request *self, const char *requestUri,
		  const char *scriptName, const char *pathInfo,
		  const char *givenScriptName, int givenScriptNameLen)
{
  int result = -1;
  char *reqUri = NULL, *r;

  if (requestUri == NULL) {
    fprintf(stderr, "REQUEST_URI is missing, cannot continue\n");
    goto out;
  }

  /* Make mutable copy (not really necessary, but...) */
  if ((reqUri = PyMem_Malloc(strlen(requestUri) + 1)) == NULL) {
    PyErr_NoMemory();
    goto out;
  }
  strcpy(reqUri, requestUri);

  /* Strip query string from REQUEST_URI */
  for (r = reqUri; *r; r++) {
    if (*r == '?') {
      *r = '\0';
      break;
    }
  }

  /* Unquote request URI */
  if (unquoteURL) {
    const char *newReqUri;
    if ((newReqUri = wsgiUnquote(reqUri)) == NULL) {
      PyErr_NoMemory();
      goto out;
    }
    PyMem_Free(reqUri);
    reqUri = (char *)newReqUri;
  }

  if (wsgiPutEnv(self, "REQUEST_URI", reqUri))
    goto out;

  if (setScriptNamePathInfo(self, reqUri, givenScriptName, givenScriptNameLen))
    goto out;

  result = 0;

 out:
  if (reqUri) PyMem_Free(reqUri);

  return result;
}

/* Only trust SCRIPT_NAME/PATH_INFO */
static int
scriptNameHandler(Request *self, const char *requestUri,
		  const char *scriptName, const char *pathInfo,
		  const char *givenScriptName, int givenScriptNameLen)
{
  int result = -1;
  int scriptNameLen;
  char *reqUri = NULL;

  if (scriptName == NULL) {
    fprintf(stderr, "SCRIPT_NAME is missing, cannot continue\n");
    goto out;
  }
  if (pathInfo == NULL) {
    pathInfo = "";
  }

  /* Rebuild REQUEST_URI from SCRIPT_NAME/PATH_INFO */
  scriptNameLen = strlen(scriptName);
  if ((reqUri = PyMem_Malloc(scriptNameLen + strlen(pathInfo) + 1))
      == NULL) {
    PyErr_NoMemory();
    goto out;
  }
  strncpy(reqUri, scriptName, scriptNameLen);
  strcpy(&reqUri[scriptNameLen], pathInfo);

  if (wsgiPutEnv(self, "REQUEST_URI", reqUri))
    goto out;

  if (setScriptNamePathInfo(self, reqUri, givenScriptName, givenScriptNameLen))
    goto out;

  result = 0;

 out:
  if (reqUri) PyMem_Free(reqUri);

  return result;
}

/* Table for env handler selection */
static struct {
  const char *name;
  envHandler handler;
} envProfiles[] = {
  { "pass-through", passThroughHandler },
  { "request-uri", requestUriHandler },
  { "script-name", scriptNameHandler }
};
#define NUM_ENV_PROFILES (sizeof(envProfiles) / sizeof(envProfiles[0]))

/* Selected env handler */
static envHandler currentEnvHandler = passThroughHandler;

/* Select an env handler by name */
int
selectEnvHandler(const char *name)
{
  int i;
  int len = strlen(name);

  if (!len) return -1; /* sanity check */

  /* Do a prefix match against profile names */
  for (i = 0; i < NUM_ENV_PROFILES; i++) {
    if (!strncasecmp(envProfiles[i].name, name, len)) {
      currentEnvHandler = envProfiles[i].handler;
      return 0;
    }
  }
  return -1;
}

/* Populate WSGI environ dictionary */
int
wsgiPopulateEnviron(Request *self)
{
  SCGIRequest *req = ((SCGIContext *)self->context)->activeReq;
  PyObject *environ = self->environ;
  PyObject *val;
  int ret, i, https = -1;
  const char *envScriptName = NULL, *scriptName;
  const char *rawReqUri = NULL, *rawScriptName = NULL, *rawPathInfo = NULL;
  int envScriptNameLen = 0, scriptNameLen;
  int result = -1;

  /* HTTP headers */
  for (i = 0; i < req->numHeaders; i++) {
    const char *header = req->headers[2 * i];
    const char *value = req->headers[2 * i + 1];

    /* Check if HTTPS */
    if (https < 0 && !strcmp("HTTPS", header)) {
      https = !strcasecmp("on", value);
    }

    /* Grab REQUEST_URI/SCRIPT_NAME/PATH_INFO */
    if (rawReqUri == NULL && !strcmp("REQUEST_URI", header)) {
      rawReqUri = value;
    }
    if (rawScriptName == NULL && !strcmp("SCRIPT_NAME", header)) {
      rawScriptName = value;
    }
    if (rawPathInfo == NULL && !strcmp("PATH_INFO", header)) {
      rawPathInfo = value;
    }

    /* Get WSGI_SCRIPT_NAME, if available */
    if (envScriptName == NULL && !strcmp("WSGI_SCRIPT_NAME", header)) {
      envScriptName = value;
      envScriptNameLen = strlen(value);
    }

    if (wsgiPutEnv(self, header, value))
      goto bad;
  }

  /* If it wasn't found, assume HTTPS is off */
  if (https < 0)
    https = 0;

  /* wsgi.version */
  if ((val = Py_BuildValue("(ii)", 1, 0)) == NULL)
    goto bad;
  ret = PyDict_SetItemString(environ, "wsgi.version", val);
  Py_DECREF(val);
  if (ret)
    goto bad;

  /* wsgi.input */
  if (PyDict_SetItemString(environ, "wsgi.input", self->input))
    goto bad;

  /* wsgi.errors */
  if (PyDict_SetItemString(environ, "wsgi.errors", wsgiStderr))
    goto bad;

  /* wsgi.multithread */
  if (PyDict_SetItemString(environ, "wsgi.multithread", multiprocess ? Py_False : Py_True))
    goto bad;

  /* wsgi.multiprocess */
  if (PyDict_SetItemString(environ, "wsgi.multiprocess", multiprocess ? Py_True : Py_False))
    goto bad;

  /* wsgi.run_once */
  if (PyDict_SetItemString(environ, "wsgi.run_once", Py_False))
    goto bad;

  /* wsgi.url_scheme */
  if (wsgiPutEnv(self, "wsgi.url_scheme", https ? "https" : "http"))
    goto bad;

  if (PyDict_SetItemString(environ, "wsgi.file_wrapper",
			   (PyObject *)&FileWrapper_Type))
    goto bad;

  /* Ensure QUERY_STRING is set */
  if (!PyDict_GetItemString(self->environ, "QUERY_STRING")) {
    if (wsgiPutEnv(self, "QUERY_STRING", ""))
      goto bad;
  }

  /* Decide which SCRIPT_NAME to use */
  scriptName = wsgiScriptName;
  scriptNameLen = wsgiScriptNameLen;
  if (!wsgiScriptNameLen && envScriptNameLen) {
    /* wsgiInit scriptName always takes priority */
    scriptName = envScriptName;
    scriptNameLen = envScriptNameLen;
  }

  /* Execute profile-specific handler */
  if (currentEnvHandler(self, rawReqUri, rawScriptName, rawPathInfo,
			scriptName, scriptNameLen))
    goto bad;

  result = 0;

 bad:
  return result;
}

/* Map SCGI errors to Python errors. */
static int
_wsgiMapError(int error)
{
   assert(error < 0);

   switch (error) {
   case SCGI_MEMORY_ERROR:
     PyErr_SetString(PyExc_MemoryError, "SCGI: Out of memory");
     break;
   case SCGI_SOCKET_ERROR:
     PyErr_SetString(PyExc_IOError, "SCGI: Socket read/write error");
     break;
   case SCGI_PROTOCOL_ERROR:
     PyErr_SetString(PyExc_ValueError, "SCGI: Protocol error");
     break;
   case SCGI_FILE_ERROR:
     PyErr_SetString(PyExc_IOError, "SCGI: File error");
     break;

   default:
     PyErr_SetString(PyExc_RuntimeError, "SCGI: Unknown error");
     break;
   }
   return -1;
}

/* Send HTTP response headers. */
int
wsgiSendHeaders(void *ctxt, int statusCode, const char *statusMsg,
		int headerCount, const char *headers[])
{
  int ret;

  Py_BEGIN_ALLOW_THREADS
  ret = scgiSendHeaders(ctxt, statusCode, statusMsg, headerCount, headers);
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

/* Send HTTP response body. */
int
wsgiSendBody(void *ctxt, const uint8_t *data, size_t len)
{
  int ret = SCGI_OK;

  Py_BEGIN_ALLOW_THREADS
  ret = scgiSendBodyChunk(ctxt, data, len);
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

/* Send a file as the HTTP response body. */
int
wsgiSendFile(void *ctxt, int fd)
{
  int ret;

  Py_BEGIN_ALLOW_THREADS
  ret = scgiSendFile(ctxt, fd);
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

/* Receive a request body chunk from the transport layer. */
int
wsgiGetBody(void *ctxt)
{
  int ret;

  Py_BEGIN_ALLOW_THREADS
  ret = scgiGetBodyChunk(ctxt);
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

/* Primes the input stream, if needed by transport layer. (SCGI sends the
   first request body chunk if Content-Length is non-zero.) */
int
wsgiPrimeInput(void *ctxt)
{
  int ret;

  if (((SCGIContext *)ctxt)->activeReq->contentLength > 0) {
    Py_BEGIN_ALLOW_THREADS
    ret = scgiProcessInput(ctxt);
    Py_END_ALLOW_THREADS

    if (ret) return _wsgiMapError(ret);
  }
  return 0;
}

/* Set application-specific (WSGI) request-level data. */
void
wsgiSetRequestData(void *ctxt, void *data)
{
  ((SCGIContext *)ctxt)->activeReq->data = data;
}

/* Get application-specific (WSGI) request-level data. */
void *
wsgiGetRequestData(void *ctxt)
{
  return ((SCGIContext *)ctxt)->activeReq->data;
}

/* Return the Content-Length associated with this request. */
int
wsgiGetContentLength(void *ctxt)
{
  return ((SCGIContext *)ctxt)->activeReq->contentLength;
}
