/*-
 * Copyright (c) 2011 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "scgi.h"

/* Request and process a body chunk from the web server. */
int
scgiGetBodyChunk(SCGIContext *ctxt)
{
  int reqLen = ctxt->activeReq->_bodyBytesAvail;
  uint8_t *oldBodyBuf;

  if (!reqLen) return SCGI_OK;

  if (reqLen > SCGI_BUFFER_SIZE)
    reqLen = SCGI_BUFFER_SIZE;

  if (fread(ctxt->bodyBuf, 1, reqLen, ctxt->sock) != reqLen)
    return SCGI_SOCKET_ERROR;

  ctxt->activeReq->_bodyBytesAvail -= reqLen;

  oldBodyBuf = ctxt->bodyBuf;
  /* Allocate new body buffer and pass old to handler */
  if ((ctxt->bodyBuf = malloc(SCGI_BUFFER_SIZE)) == NULL)
    return SCGI_MEMORY_ERROR;
  return ctxt->bodyHandler(ctxt, oldBodyBuf, reqLen, oldBodyBuf);
}

#if !defined(HAVE_FREEBSD_SENDFILE) && !defined(HAVE_DARWIN_SENDFILE)

/* Dumps a file as a stream of SEND_BODY_CHUNK packets.
   Non-sendfile(2) version. */
int
scgiSendFile(SCGIContext *ctxt, int fd)
{
  size_t readLen;

  do {
    readLen = fread(ctxt->bodyBuf, 1, SCGI_BUFFER_SIZE, ctxt->sock);
    if (ferror(ctxt->sock))
      return SCGI_FILE_ERROR;

    if (readLen > 0) {
      if (fwrite(ctxt->bodyBuf, 1, readLen, ctxt->sock) != readLen)
	return SCGI_SOCKET_ERROR;
    }
  }
  while (readLen == SCGI_BUFFER_SIZE);

  if (fflush(ctxt->sock))
    return SCGI_SOCKET_ERROR;

  return SCGI_OK;
}

#else /* !HAVE_FREEBSD_SENDFILE && !HAVE_DARWIN_SENDFILE */

/* Dumps a file as a stream of SEND_BODY_CHUNK packets.
   FreeBSD sendfile(2) version. */
int
scgiSendFile(SCGIContext *ctxt, int fd)
{
  off_t start;
#ifdef HAVE_DARWIN_SENDFILE
  off_t size;
#endif /* HAVE_DARWIN_SENDFILE */

  if ((start = lseek(fd, 0, SEEK_CUR)) < 0)
    return SCGI_FILE_ERROR;

  /* Flush before we operate directly on the socket */
  if (fflush(ctxt->sock))
    return SCGI_SOCKET_ERROR;

#ifdef HAVE_FREEBSD_SENDFILE
  if (sendfile(fd, fileno(ctxt->sock), start, 0, NULL, NULL, 0))
    return SCGI_FILE_ERROR;
#else /* HAVE_FREEBSD_SENDFILE */
  size = 0; /* send til EOF */
  if (sendfile(fd, fileno(ctxt->sock), start, &size, NULL, 0))
    return SCGI_FILE_ERROR;
#endif /* HAVE_FREEBSD_SENDFILE */

  return SCGI_OK;
}

#endif /* !HAVE_FREEBSD_SENDFILE */

/* Send a response body chunk. The data may be safely recycled upon return
   of this function. */
int
scgiSendBodyChunk(SCGIContext *ctxt, const uint8_t *data, size_t len)
{
  if (fwrite(data, 1, len, ctxt->sock) != len)
    return SCGI_SOCKET_ERROR;

  if (fflush(ctxt->sock))
    return SCGI_SOCKET_ERROR;

  return SCGI_OK;
}

/* Begin a response and send headers. All string arguments are copied. */
int
scgiSendHeaders(SCGIContext *ctxt,
		uint16_t statusCode, const char *statusMsg,
		int headerCount, const char **headers)
{
  int i;

  if (fprintf(ctxt->sock, "Status: %d\r\n", statusCode) < 0)
    return SCGI_SOCKET_ERROR;

  for (i = 0; i < headerCount; i++) {
    int j = 2 * i;
    if (fprintf(ctxt->sock, "%s: %s\r\n", headers[j], headers[j + 1]) < 0)
      return SCGI_SOCKET_ERROR;
  }

  if (fprintf(ctxt->sock, "\r\n") < 0)
    return SCGI_SOCKET_ERROR;

  return SCGI_OK;
}

/* Main loop. Waits for and processes a single request. If called within the
   context of an active request, forwards to scgiGetBodyChunk() instead. */
int
scgiProcessInput(SCGIContext *ctxt)
{
  int error = SCGI_ERROR;
  SCGIRequest *req;
  unsigned long len, i;
  char *p;
  int j;

  /* If request already active, grab a body chunk instead */
  if (ctxt->activeReq != NULL)
    return scgiGetBodyChunk(ctxt);

  /* Allocate request */
  if ((req = calloc(1, sizeof(*req))) == NULL) {
    error = SCGI_MEMORY_ERROR;
    goto bad;
  }

  /* Decode header netstring */
  if (fscanf(ctxt->sock, "%9lu", &len) < 1) {
    error = SCGI_PROTOCOL_ERROR;
    goto bad;
  }
    
  if (fgetc(ctxt->sock) != ':') {
    error = SCGI_PROTOCOL_ERROR;
    goto bad;
  }

  if ((req->rawHeaders = malloc(len + 1)) == NULL) {
    error = SCGI_MEMORY_ERROR;
    goto bad;
  }

  if (fread(req->rawHeaders, 1, len, ctxt->sock) != len) {
    error = SCGI_PROTOCOL_ERROR;
    goto bad;
  }

  if (fgetc(ctxt->sock) != ',') {
    error = SCGI_PROTOCOL_ERROR;
    goto bad;
  }

  /* Count the number of headers (can't know in advance... too bad) */
  for (i = 0, p = req->rawHeaders; i < len;) {
    /* Skip header name */
    while (i++ < len && *(p++));
    /* Skip header value */
    while (i++ < len && *(p++));

    /* Got one complete header, count it */
    if (i <= len)
      req->numHeaders++;
  }

  /* Now put header/value pairs into array */
  if ((req->headers = calloc(sizeof(*req->headers),
			     req->numHeaders * 2)) == NULL) {
    error = SCGI_MEMORY_ERROR;
    goto bad;
  }

  req->contentLength = -1;

  for (i = 0, p = req->rawHeaders, j = 0; i < len;) {
    const char *name, *value;

    name = p;
    while (i++ < len && *(p++));
    value = p;
    while (i++ < len && *(p++));

    if (i <= len) {
      req->headers[j++] = name;
      req->headers[j++] = value;

      /* Special handling of Content-Length header */
      if (req->contentLength < 0 && !strcmp(name, "CONTENT_LENGTH")) {
	req->contentLength = strtol(value, NULL, 10);
	if (req->contentLength < 0) /* do this now to skip this block */
	  req->contentLength = 0;
      }
    }
  }

  if (req->contentLength < 0)
    req->contentLength = 0;

  ctxt->activeReq = req;

  req->_bodyBytesAvail = req->contentLength;

  /* Execute handler */
  error = ctxt->handler(ctxt);
  if (error == SCGI_OK)
    error = SCGI_DONE;

  /* Fallthrough */

 bad:
  ctxt->activeReq = NULL;

  if (req != NULL) {
    if (req->headers != NULL) free(req->headers);
    if (req->rawHeaders != NULL) free(req->rawHeaders);
    free(req);
  }

  return error;
}

/* Frees an SCGI context. */
void
scgiDestroyContext(SCGIContext *ctxt)
{
  if (ctxt != NULL) {
    if (ctxt->bodyBuf) free(ctxt->bodyBuf);
    if (ctxt->sock != NULL) fclose(ctxt->sock);
    free(ctxt);
  }
}

/* Dummy body handler */
static int
nullBodyHandler(SCGIContext *ctxt, uint8_t *data, size_t len, void *tofree)
{
  if (tofree != NULL) free(tofree);
  return SCGI_OK;
}

/* Allocates and initializes an SCGIContext. Each thread must have its own
   SCGIContext. The passed-in socket fd should not be closed by the caller --
   it will be closed when the SCGIContext is destroyed. */
SCGIContext *
scgiCreateContext(int sock, scgi_handler_t handler,
		  scgi_body_handler_t bodyHandler)
{
  SCGIContext *ctxt = NULL;

  if ((ctxt = calloc(1, sizeof(*ctxt))) == NULL) {
    close(sock);
    goto bad;
  }

  if ((ctxt->sock = fdopen(sock, "w+b")) == NULL) {
    close(sock);
    goto bad;
  }

  ctxt->handler = handler;
  ctxt->bodyHandler = bodyHandler ? bodyHandler : nullBodyHandler;

  if ((ctxt->bodyBuf = malloc(SCGI_BUFFER_SIZE)) == NULL)
    goto bad;

  return ctxt;

 bad:
  scgiDestroyContext(ctxt);

  return NULL;
}
