/*-
 * Copyright (c) 2011 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include <stdlib.h>

#include <pthread.h>

#include "threadpool.h"

/* Worker thread main loop */
static void *
thread_main(void *data)
{
  ThreadPoolContext *ctxt = data;
  int eternal;

  pthread_mutex_lock(&ctxt->contextMutex);

  /* If under the minimum, flag as eternal */
  eternal = ctxt->numThreads < ctxt->minThreads;

  ctxt->numThreads++;

  for(;;) {
    struct _ThreadPoolJob *job;
    int timedout = 0;

    /* Wait for a job */
    while (ctxt->jobQueueHead.next->next == NULL) {
      if (eternal) {
	pthread_cond_wait(&ctxt->jobQueueCond, &ctxt->contextMutex);
      }
      else {
	struct timespec timeout;

	timeout.tv_sec = ctxt->timeToIdle;
	timeout.tv_nsec = 0;
	if (pthread_cond_timedwait(&ctxt->jobQueueCond, &ctxt->contextMutex,
				   &timeout)) {
	  timedout = 1;
	  break;
	}
      }
    }

    /* NB contextMutex re-acquired */

    if (timedout) break;

    /* Dequeue next job */
    job = ctxt->jobQueueHead.next;
    if (job->next == NULL) continue; /* sanity check */
    job->prev->next = job->next;
    job->next->prev = job->prev;

    /* Indicate we're busy */
    ctxt->busyThreads++;

    pthread_mutex_unlock(&ctxt->contextMutex);

    /* Run job */
    job->handler(job->data);

    /* All done */
    pthread_mutex_lock(&ctxt->contextMutex);

    ctxt->busyThreads--;
  }

  ctxt->numThreads--;

  pthread_mutex_unlock(&ctxt->contextMutex);

  return NULL;
}

/* Create a new worker thread */
static int
new_thread(ThreadPoolContext *ctxt)
{
  pthread_t thread;

  if (!pthread_create(&thread, NULL, thread_main, ctxt)) {
    pthread_detach(thread);
    return 0;
  }

  return -1;
}

int
threadpool_run(ThreadPoolContext *ctxt, threadpool_handler_t handler,
	       void *data)
{
  int result = 0;
  struct _ThreadPoolJob *job;

  /* Allocate new job */
  if ((job = malloc(sizeof(*job))) != NULL) {
    job->handler = handler;
    job->data = data;

    pthread_mutex_lock(&ctxt->contextMutex);

    /* Queue it up */
    job->next = &ctxt->jobQueueTail;
    job->prev = ctxt->jobQueueTail.prev;

    ctxt->jobQueueTail.prev->next = job;
    ctxt->jobQueueTail.prev = job;

    /* Is a free thread available? */
    if (ctxt->busyThreads >= ctxt->numThreads &&
	ctxt->numThreads < ctxt->maxThreads) {
      /* If not, and we're still under the limit, create one */
      result = new_thread(ctxt);
    }

    /* Signal a worker thread */
    pthread_cond_signal(&ctxt->jobQueueCond);

    pthread_mutex_unlock(&ctxt->contextMutex);

    return result;
  }

  return -1;
}

void
threadpool_shutdown(ThreadPoolContext *ctxt)
{
  if (ctxt) {
    /* TODO shutdown worker threads */
    pthread_cond_destroy(&ctxt->jobQueueCond);
    pthread_mutex_destroy(&ctxt->contextMutex);
    free(ctxt);
  }
}

ThreadPoolContext *
threadpool_init(int minThreads, int maxThreads, time_t timeToIdle)
{
  ThreadPoolContext *ctxt;
  int i;

  if ((ctxt = calloc(1, sizeof(*ctxt))) == NULL)
    return NULL;

  ctxt->minThreads = minThreads;
  ctxt->maxThreads = maxThreads;
  ctxt->timeToIdle = timeToIdle;

  if (pthread_mutex_init(&ctxt->contextMutex, NULL))
    goto bad;

  if (pthread_cond_init(&ctxt->jobQueueCond, NULL)) {
    pthread_mutex_destroy(&ctxt->contextMutex);
    goto bad;
  }

  /* Initialize job queue */
  ctxt->jobQueueHead.next = &ctxt->jobQueueTail;
  ctxt->jobQueueHead.prev = NULL;

  ctxt->jobQueueTail.next = NULL;
  ctxt->jobQueueTail.prev = &ctxt->jobQueueHead;

  /* Fully initialized at this point */

  /* Create minimum number of threads */
  for (i = 0; i < ctxt->minThreads; i++) {
    if (new_thread(ctxt)) {
      threadpool_shutdown(ctxt);
      return NULL;
    }
  }

  return ctxt;

 bad:
  free(ctxt);
  return NULL;
}
